@@ -19,6 +19,8 @@ from .config import (
     check_prices,
     check_server_type,
     check_ssh_key,
+    check_setup_script,
+    check_startup_script,
 )
 from .config import standby_runner
 from .config import read, write
