@@ -6,6 +6,7 @@ import base64
 import hashlib
 import logging
 import logging.config
+import dataclasses
 
 from dataclasses import dataclass
 
@@ -31,7 +32,12 @@ default_user_config = os.path.expanduser("~/.github-hetzner-runners/config.yaml"
 def env_constructor(loader, node):
     value = loader.construct_scalar(node)
     for group in env_pattern.findall(value):
-        value = value.replace(f"${{{group}}}", os.environ.get(group))
+        env_value = os.environ.get(group)
+        if env_value is None:
+            assert (
+                False
+            ), f"environment variable ${group} used in the config is not defined"
+        value = value.replace(f"${{{group}}}", env_value)
     return value
 
 
@@ -47,6 +53,14 @@ class ImageError(Exception):
     pass
 
 
+class SetupScriptError(Exception):
+    pass
+
+
+class StartupScriptError(Exception):
+    pass
+
+
 class ServerTypeError(Exception):
     pass
 
@@ -61,6 +75,7 @@ image = args.image_type
 location = args.location_type
 server_type = args.server_type
 end_of_life = args.end_of_life_type
+meta_label_type = args.meta_label_type
 
 
 @dataclass
@@ -71,7 +86,7 @@ class standby_runner:
 
 
 @dataclass
-class deploy:
+class deploy_:
     server_type: ServerType = server_type("cpx11")
     image: Image = image("x86:system:ubuntu-22.04")
     location: Location = None
@@ -81,7 +96,7 @@ class deploy:
 @dataclass
 class cloud:
     server_name: str = "github-hetzner-runners"
-    deploy: deploy = deploy()
+    deploy: deploy_ = dataclasses.field(default_factory=deploy_)
 
 
 @dataclass
@@ -93,24 +108,19 @@ class Config:
     hetzner_token: str = os.getenv("HETZNER_TOKEN")
     ssh_key: str = os.path.expanduser("~/.ssh/id_rsa.pub")
     additional_ssh_keys: list[str] = None
-    with_label: str = None
+    with_label: list[str] = None
     label_prefix: str = ""
+    meta_label: dict[str, set[str]] = None
     recycle: bool = True
     end_of_life: int = 50
     delete_random: bool = False
     max_runners: int = 10
     max_runners_in_workflow_run: int = None
     default_image: Image = image("x86:system:ubuntu-22.04")
-    default_server_type: ServerType = server_type("cx11")
+    default_server_type: ServerType = server_type("cx22")
     default_location: Location = None
     workers: int = 10
-    setup_script: str = os.path.join(current_dir, "..", "scripts", "setup.sh")
-    startup_x64_script: str = os.path.join(
-        current_dir, "..", "scripts", "startup_x64.sh"
-    )
-    startup_arm64_script: str = os.path.join(
-        current_dir, "..", "scripts", "startup_arm64.sh"
-    )
+    scripts: str = os.path.join(current_dir, "..", "scripts")
     max_powered_off_time: int = 60
     max_unused_runner_time: int = 180
     max_runner_registration_time: int = 180
@@ -123,18 +133,24 @@ class Config:
     embedded_mode: bool = False
     logger_config: dict = None
     logger_format: dict = None
-    cloud: cloud = cloud()
+    cloud: cloud = dataclasses.field(default_factory=cloud)
     standby_runners: list[standby_runner] = None
-    server_prices: dict[str, float] = None
+    server_prices: dict[str, dict[str, float]] = None
     config_file: str = None
 
     def __post_init__(self):
+        if self.with_label is None:
+            self.with_label = ["self-hosted"]
+
         if self.standby_runners is None:
             self.standby_runners = []
 
         if self.additional_ssh_keys is None:
             self.additional_ssh_keys = []
 
+        if self.meta_label is None:
+            self.meta_label = {}
+
         if self.logger_format is None:
             self.logger_format = logger_format
 
@@ -218,6 +234,21 @@ def parse_config(filename: str):
 
     doc = doc["config"]
 
+    if doc.get("setup_script"):
+        assert (
+            False
+        ), "config.setup_script is deprecated, use the new config.scripts option"
+
+    if doc.get("startup_x64_script"):
+        assert (
+            False
+        ), "config.startup_x64_script is deprecated, use the new config.scripts option"
+
+    if doc.get("startup_arm64_script"):
+        assert (
+            False
+        ), "config.startup_x64_script is deprecated, see the new config.scripts option"
+
     if doc.get("ssh_key") is not None:
         assert isinstance(doc["ssh_key"], str), "config.ssh_key: is not a string"
         doc["ssh_key"] = path(doc["ssh_key"], check_exists=False)
@@ -232,13 +263,32 @@ def parse_config(filename: str):
             ), f"config.additional_ssh_keys[{i}]: is not a string"
 
     if doc.get("with_label") is not None:
-        assert isinstance(doc["with_label"], str), "config.with_label: is not a string"
+        assert isinstance(doc["with_label"], list), "config.with_label: is not a list"
+        for i, label in enumerate(doc["with_label"]):
+            assert isinstance(label, str), f"config.with_label[{i}]: is not a string"
 
     if doc.get("label_prefix") is not None:
         assert isinstance(
             doc["label_prefix"], str
         ), "config.label_prefix: is not a string"
 
+    if doc.get("meta_label") is not None:
+        assert isinstance(
+            doc["meta_label"], dict
+        ), "config.meta_label is not a dictionary"
+        for i, meta in enumerate(doc["meta_label"]):
+            assert isinstance(
+                meta, str
+            ), f"config.meta_label.{meta}: name is not a string"
+            assert isinstance(
+                doc["meta_label"][meta], list
+            ), f"config.meta_label.{meta}: is not a list"
+            for j, v in enumerate(doc["meta_label"][meta]):
+                assert isinstance(
+                    v, str
+                ), f"config.meta_label.{meta}[{j}]: is not a string"
+            doc["meta_label"][meta] = set(doc["meta_label"][meta])
+
     if doc.get("recycle") is not None:
         assert isinstance(doc["recycle"], bool), "config.recycle: is not a boolean"
 
@@ -286,23 +336,11 @@ def parse_config(filename: str):
         v = doc["workers"]
         assert isinstance(v, int) and v > 0, "config.workers: is not an integer > 0"
 
-    if doc.get("setup_script") is not None:
+    if doc.get("scripts") is not None:
         try:
-            doc["setup_script"] = path(doc["setup_script"])
+            doc["scripts"] = path(doc["scripts"])
         except Exception as e:
-            assert False, f"config.setup_script: {e}"
-
-    if doc.get("startup_x64_script") is not None:
-        try:
-            doc["startup_x64_script"] = path(doc["startup_x64_script"])
-        except Exception as e:
-            assert False, f"config.startup_x64_script: {e}"
-
-    if doc.get("startup_arm64_script") is not None:
-        try:
-            doc["startup_arm64_script"] = path(doc["startup_arm64_script"])
-        except Exception as e:
-            assert False, f"config.startup_arm64_script: {e}"
+            assert False, f"config.scripts: {e}"
 
     if doc.get("max_powered_off_time") is not None:
         v = doc["max_powered_off_time"]
@@ -350,7 +388,7 @@ def parse_config(filename: str):
         assert (
             doc["logger_config"]["loggers"].get("testflows.github.hetzner.runners")
             is not None
-        ), 'config.logger_config.loggers."tesflows.github.hetzner.runners" is not defined'
+        ), 'config.logger_config.loggers."testflows.github.hetzner.runners" is not defined'
         assert (
             doc["logger_config"]["loggers"]["testflows.github.hetzner.runners"].get(
                 "handlers"
@@ -483,11 +521,11 @@ def parse_config(filename: str):
         if doc["cloud"].get("server_name"):
             doc["cloud"] = cloud(
                 doc["cloud"]["server_name"],
-                deploy=deploy(**doc["cloud"].get("deploy", {})),
+                deploy=deploy_(**doc["cloud"].get("deploy", {})),
             )
         else:
             doc["cloud"] = cloud(
-                deploy=deploy(**doc["cloud"].get("deploy", {})),
+                deploy=deploy_(**doc["cloud"].get("deploy", {})),
             )
 
     if doc.get("standby_runners"):
@@ -560,12 +598,20 @@ def check_ssh_key(client: Client, ssh_key: str, is_file=True):
         name=name, public_key=public_key, fingerprint=fingerprint(public_key)
     )
 
-    if not client.ssh_keys.get_by_fingerprint(fingerprint=ssh_key.fingerprint):
+    existing_ssh_key = client.ssh_keys.get_by_fingerprint(
+        fingerprint=ssh_key.fingerprint
+    )
+
+    if not existing_ssh_key:
         with Action(
             f"Creating SSH key {ssh_key.name} with fingerprint {ssh_key.fingerprint}",
             stacklevel=3,
         ):
-            client.ssh_keys.create(name=ssh_key.name, public_key=ssh_key.public_key)
+            ssh_key = client.ssh_keys.create(
+                name=ssh_key.name, public_key=ssh_key.public_key
+            )
+    else:
+        ssh_key = existing_ssh_key
 
     return ssh_key
 
@@ -619,6 +665,7 @@ def check_server_type(client: Client, server_type: ServerType):
 
 
 def check_prices(client: Client):
+    """Check server prices."""
     server_types: list[ServerType] = client.server_types.get_all()
     return {
         t.name.lower(): {
@@ -627,3 +674,17 @@ def check_prices(client: Client):
         }
         for t in server_types
     }
+
+
+def check_setup_script(script: str):
+    """Check if setup script is valid."""
+    if not os.path.exists(script):
+        raise SetupScriptError(f"invalid setup script path '{script}'")
+    return script
+
+
+def check_startup_script(script: str):
+    """Check if startup script is valid."""
+    if not os.path.exists(script):
+        raise StartupScriptError(f"invalid startup script path '{script}'")
+    return script
