@@ -13,10 +13,8 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 import os
-import math
 import time
-import uuid
-import random
+import queue
 import logging
 import threading
 
@@ -26,12 +24,12 @@ from .actions import Action
 from .request import request
 from .args import image_type
 from .logger import logger
-from .config import Config, check_image
+from .config import Config, check_image, check_startup_script, check_setup_script
 from .config import standby_runner as StandbyRunner
 
-from .server import wait_ssh, ssh, wait_ready, age
+from .server import wait_ssh, ssh
 
-from hcloud import Client
+from hcloud import Client, APIException
 from hcloud.ssh_keys.domain import SSHKey
 from hcloud.server_types.domain import ServerType
 from hcloud.locations.domain import Location
@@ -42,7 +40,6 @@ from hcloud.helpers.labels import LabelValidator
 
 from github import Github
 from github.Repository import Repository
-from github.WorkflowJob import WorkflowJob
 from github.WorkflowRun import WorkflowRun
 from github.SelfHostedActionsRunner import SelfHostedActionsRunner
 
@@ -56,6 +53,23 @@ recycle_server_name_prefix = f"{server_name_prefix}recycle-"
 server_ssh_key_label = "github-hetzner-runner-ssh-key"
 
 
+@dataclass
+class ScaleUpFailureMessage:
+    """Scale up server failure."""
+
+    time: float
+    labels: set[str]
+    server_name: str
+    exception: Exception
+
+
+class MaxNumberOfServersReached(Exception):
+    """Exception to indicate that scale up service
+    reached maximum number of servers."""
+
+    pass
+
+
 @dataclass
 class RunnerServer:
     name: str
@@ -69,7 +83,23 @@ class RunnerServer:
 
 def uid():
     """Return unique id - just a timestamp."""
-    return str(time.time())
+    return str(time.time()).replace(".", "")
+
+
+def get_runner_server_name(runner_name: str) -> str:
+    """Determine runner's server name."""
+    return "-".join(runner_name.split("-")[:5])
+
+
+def get_runner_server_type_and_location(runner_name: str):
+    """Determine runner's server type, and location."""
+    server_type, server_location = None, None
+
+    if runner_name and runner_name.startswith(runner_name_prefix):
+        if len(runner_name.split("-")) == 7:
+            server_type, server_location = runner_name.split("-")[5:]
+
+    return server_type, server_location
 
 
 def server_setup(
@@ -98,9 +128,6 @@ def server_setup(
         )
         GITHUB_RUNNER_TOKEN = content["token"]
 
-    with Action("Getting current directory", server_name=server.name):
-        current_dir = os.path.dirname(__file__)
-
     with Action("Executing setup.sh script", server_name=server.name):
         ssh(server, f"bash -s  < {setup_script}", stacklevel=5)
 
@@ -108,10 +135,13 @@ def server_setup(
         ssh(
             server,
             f"'sudo -u ubuntu "
-            f"GITHUB_REPOSITORY=\"{os.getenv('GITHUB_REPOSITORY')}\" "
+            f'GITHUB_REPOSITORY="{github_repository}" '
             f'GITHUB_RUNNER_TOKEN="{GITHUB_RUNNER_TOKEN}" '
             f"GITHUB_RUNNER_GROUP=Default "
             f'GITHUB_RUNNER_LABELS="{runner_labels}" '
+            f'SERVER_ID="{server.id}" '
+            f'SERVER_TYPE_NAME="{server.server_type.name}" '
+            f'SERVER_LOCATION_NAME="{server.datacenter.location.name}" '
             f"bash -s' < {startup_script}",
             stacklevel=5,
         )
@@ -125,11 +155,10 @@ def get_server_type(labels: set[str], default: ServerType, label_prefix: str = "
         label_prefix += "-"
     label_prefix += "type-"
 
-    if server_type is None:
-        for label in labels:
-            if label.startswith(label_prefix):
-                server_type_name = label.split(label_prefix, 1)[-1].lower()
-                server_type = ServerType(name=server_type_name)
+    for label in labels:
+        if label.startswith(label_prefix):
+            server_type_name = label.split(label_prefix, 1)[-1].lower()
+            server_type = ServerType(name=server_type_name)
 
     if server_type is None:
         server_type = default
@@ -151,11 +180,10 @@ def get_server_location(
         label_prefix += "-"
     label_prefix += "in-"
 
-    if server_location is None:
-        for label in labels:
-            if label.startswith(label_prefix):
-                server_location_name = label.split(label_prefix, 1)[-1].lower()
-                server_location = Location(name=server_location_name)
+    for label in labels:
+        if label.startswith(label_prefix):
+            server_location_name = label.split(label_prefix, 1)[-1].lower()
+            server_location = Location(name=server_location_name)
 
     if server_location is None and default:
         server_location = default
@@ -173,13 +201,12 @@ def get_server_image(
         label_prefix += "-"
     label_prefix += "image-"
 
-    if server_image is None:
-        for label in labels:
-            if label.startswith(label_prefix):
-                server_image = check_image(
-                    client,
-                    image_type(label.split(label_prefix, 1)[-1].lower(), separator="-"),
-                )
+    for label in labels:
+        if label.startswith(label_prefix):
+            server_image = check_image(
+                client,
+                image_type(label.split(label_prefix, 1)[-1].lower(), separator="-"),
+            )
 
     if server_image is None:
         server_image = default
@@ -187,15 +214,89 @@ def get_server_image(
     return server_image
 
 
-def get_startup_script(server_type: ServerType, x64, arm64):
-    """Get startup script based on the requested server type.
+def get_server_arch(server_type: ServerType):
+    """Get server architecture base on the requested server type.
     ARM64 servers type names start with "CA" prefix.
 
     For example, CAX11, CAX21, CAX31, and CAX41
     """
     if server_type.name.lower().startswith("ca"):
-        return arm64
-    return x64
+        return "arm64"
+    return "x64"
+
+
+def get_setup_script(
+    scripts: str, labels: set[str], default: str = "setup.sh", label_prefix: str = ""
+):
+    """Get setup script."""
+    script = None
+
+    if label_prefix and not label_prefix.endswith("-"):
+        label_prefix += "-"
+    label_prefix += "setup-"
+
+    for label in labels:
+        if label.startswith(label_prefix):
+            script = label.split(label_prefix, 1)[-1] + ".sh"
+
+    if script is None:
+        script = default
+
+    script = check_setup_script(os.path.join(scripts, script))
+
+    return script
+
+
+def get_startup_script(
+    scripts: str,
+    server_type: ServerType,
+    labels: set[str],
+    default: str = "startup-{arch}.sh",
+    label_prefix: str = "",
+):
+    """Get startup script based on the requested server type.
+    ARM64 servers type names start with "CA" prefix.
+
+    For example, CAX11, CAX21, CAX31, and CAX41
+    """
+    script = None
+    default = default.format(arch=get_server_arch(server_type))
+
+    if label_prefix and not label_prefix.endswith("-"):
+        label_prefix += "-"
+    label_prefix += "startup-"
+
+    for label in labels:
+        if label.startswith(label_prefix):
+            script = label.split(label_prefix, 1)[-1] + ".sh"
+
+    if script is None:
+        script = default
+
+    script = check_startup_script(os.path.join(scripts, script))
+
+    return script
+
+
+def expand_meta_label(
+    meta_label: dict[str, set[str]], labels: set[str], label_prefix: str = ""
+):
+    """Expand any meta labels."""
+    expanded_labels = []
+
+    for label in labels:
+        expanded_labels.append(label)
+        if label.startswith(label_prefix):
+            raw_label = label.split(label_prefix, 1)[-1] if label_prefix else label
+            if raw_label in meta_label:
+                expanded_labels += list(meta_label[label])
+
+    return set(expanded_labels)
+
+
+def raise_exception(exc):
+    """Task to raise an exception using the worker pool."""
+    raise exc
 
 
 def create_server(
@@ -237,11 +338,6 @@ def create_server(
         )
         server: BoundServer = response.server
 
-    with Action(
-        f"Waiting for server {server.name} to be ready", server_name=name
-    ) as action:
-        wait_ready(server=server, timeout=timeout, action=action)
-
     setup_worker_pool.submit(
         server_setup,
         server=response.server,
@@ -290,11 +386,6 @@ def recycle_server(
     with Action(f"Rebuilding recycled server {server.name} image", server_name=name):
         server.rebuild(image=server_image).wait_until_finished(max_retries=timeout)
 
-    with Action(
-        f"Waiting for server {server.name} to be ready", server_name=name
-    ) as action:
-        wait_ready(server=server, timeout=timeout, action=action)
-
     setup_worker_pool.submit(
         server_setup,
         server=server,
@@ -307,10 +398,22 @@ def recycle_server(
     )
 
 
+def count_available_runners(runners: list[SelfHostedActionsRunner], labels: set[str]):
+    """Return number of available runners that match labels (subset)."""
+    count = 0
+
+    for runner in runners:
+        if runner.status == "online":
+            runner_labels = set([label["name"] for label in runner.labels()])
+            if labels.issubset(runner_labels):
+                if not runner.busy:
+                    count += 1
+
+    return count
+
+
 def count_available(servers: list[RunnerServer], labels: set[str]):
-    """Return number of available servers that match
-    that match labels (subset).
-    """
+    """Return number of available servers that match labels (subset)."""
     count = 0
 
     for runner_server in servers:
@@ -324,9 +427,7 @@ def count_available(servers: list[RunnerServer], labels: set[str]):
 
 
 def count_present(servers: list[RunnerServer], labels: set[str]):
-    """Return number of present servers that match
-    that match labels (subset).
-    """
+    """Return number of present servers that match labels (subset)."""
     count = 0
 
     for runner_server in servers:
@@ -388,6 +489,7 @@ def recyclable_server_match(
 
 def scale_up(
     terminate: threading.Event,
+    mailbox: queue.Queue,
     worker_pool: ThreadPoolExecutor,
     ssh_keys: list[SSHKey],
     config: Config,
@@ -406,9 +508,10 @@ def scale_up(
     debug: bool = config.debug
     standby_runners: list[StandbyRunner] = config.standby_runners
     recycle: bool = config.recycle
-    server_prices: dict[str, float] = config.server_prices
-    with_label: str = config.with_label
+    with_label: list[str] = config.with_label
     label_prefix: str = config.label_prefix
+    meta_label: dict[str, set[str]] = config.meta_label
+    scripts: str = config.scripts
     interval: int = -1
 
     with Action("Logging in to Hetzner Cloud"):
@@ -424,6 +527,8 @@ def scale_up(
         """Create new server that would provide a runner with given labels."""
         recyclable_servers: list[BoundServer] = []
 
+        labels = expand_meta_label(meta_label=meta_label, labels=labels)
+
         server_type = get_server_type(
             labels=labels, default=default_server_type, label_prefix=label_prefix
         )
@@ -436,11 +541,16 @@ def scale_up(
             default=default_image,
             label_prefix=label_prefix,
         )
-        setup_script = config.setup_script
+        setup_script = get_setup_script(
+            scripts=scripts,
+            labels=labels,
+            label_prefix=label_prefix,
+        )
         startup_script = get_startup_script(
+            scripts=scripts,
             server_type=server_type,
-            x64=config.startup_x64_script,
-            arm64=config.startup_arm64_script,
+            labels=labels,
+            label_prefix=label_prefix,
         )
 
         with Action(
@@ -485,6 +595,7 @@ def scale_up(
                             timeout=max_server_ready_time,
                         )
                         future.server_name = name
+                        future.server_labels = labels
                         futures.append(future)
                         servers.pop(servers.index(server))
                         servers.append(
@@ -512,45 +623,16 @@ def scale_up(
                     stacklevel=3,
                     server_name=name,
                 ):
-                    if recyclable_servers:
-                        picking = "randomly picked"
-                        if server_prices is None:
-                            random.shuffle(recyclable_servers)
-                        else:
-                            picking = "the cheapest"
-
-                            def sorting_key(server):
-                                server_type_name = server.server_type.name
-                                server_location_name = server.server_location.name
-                                server_age = age(server.server) if server.server else 0
-                                try:
-                                    return (60 - server_age.minutes) - server_prices[
-                                        server_type_name
-                                    ][server_location_name] / 60
-                                except KeyError:
-                                    with Action(
-                                        f"price for {server_type_name} at {server_location_name} is missing",
-                                        level=logging.ERROR,
-                                        stacklevel=4,
-                                        server_name=name,
-                                    ):
-                                        return math.inf
-
-                            recyclable_servers.sort(key=sorting_key, reverse=True)
-
-                        recyclable_server = recyclable_servers.pop()
-
-                        with Action(
-                            f"Deleting {picking} recyclable server {recyclable_server.name} with type "
-                            f"{recyclable_server.server_type.name}",
-                            stacklevel=3,
-                            ignore_fail=True,
-                            server_name=name,
-                        ):
-                            recyclable_server.server.delete()
-                        servers.pop(servers.index(recyclable_server))
-                    else:
-                        raise StopIteration("maximum number of servers reached")
+                    future = worker_pool.submit(
+                        raise_exception,
+                        exc=MaxNumberOfServersReached(
+                            f"maximum number of servers reached {len(servers)}/{max_servers}"
+                        ),
+                    )
+                    future.server_name = name
+                    future.server_labels = labels
+                    futures.append(future)
+                    raise StopIteration("maximum number of servers reached")
 
         future = worker_pool.submit(
             create_server,
@@ -569,6 +651,7 @@ def scale_up(
             timeout=max_server_ready_time,
         )
         future.server_name = name
+        future.server_labels = labels
         futures.append(future)
         servers.append(
             RunnerServer(
@@ -588,7 +671,6 @@ def scale_up(
     with ThreadPoolExecutor(
         max_workers=worker_pool._max_workers, thread_name_prefix=f"setup-worker"
     ) as setup_worker_pool:
-
         while True:
             interval += 1
             if terminate.is_set():
@@ -645,7 +727,7 @@ def scale_up(
                 ):
                     for runner in runners:
                         for server in servers:
-                            if server.name == runner.name:
+                            if runner.name.startswith(server.name):
                                 if runner.status == "online":
                                     server.status = "busy" if runner.busy else "ready"
 
@@ -700,6 +782,12 @@ def scale_up(
                                             continue
 
                                     if job.status == "in_progress":
+                                        # skip jobs that were assigned to some other runners
+                                        if not job.raw_data["runner_name"].startswith(
+                                            runner_name_prefix
+                                        ):
+                                            continue
+
                                         # check if the job is running on a standby runner
                                         if job.raw_data["runner_name"].startswith(
                                             standby_runner_name_prefix
@@ -730,11 +818,28 @@ def scale_up(
                                             break
 
                                     if with_label is not None:
-                                        if not with_label in labels:
+                                        found_all_with_labels = True
+                                        for label in with_label:
+                                            if not label in labels:
+                                                found_all_with_labels = False
+                                                with Action(
+                                                    f"Skipping {job} with {labels} as it is missing label '{label}'",
+                                                    server_name=server_name,
+                                                    interval=interval,
+                                                ):
+                                                    break
+                                        if not found_all_with_labels:
+                                            continue
+
+                                    with Action(
+                                        f"Checking available runners for {job}"
+                                    ):
+                                        available = count_available_runners(
+                                            runners=runners, labels=labels
+                                        )
+                                        if available > 0:
                                             with Action(
-                                                f"Skipping {job} with {labels} as it is missing label '{with_label}'",
-                                                server_name=server_name,
-                                                interval=interval,
+                                                f"Found at least one potentially available runner for {job}"
                                             ):
                                                 continue
 
@@ -763,7 +868,9 @@ def scale_up(
                                     servers=servers, labels=labels
                                 )
                             else:
-                                available = count_present(server=servers, labels=labels)
+                                available = count_present(
+                                    servers=servers, labels=labels
+                                )
 
                             if available < standby_runner.count:
                                 for _ in range(standby_runner.count - available):
@@ -791,7 +898,35 @@ def scale_up(
                         server_name=future.server_name,
                         interval=interval,
                     ):
-                        future.result()
+                        try:
+                            future.result()
+                        except Exception as exc:
+                            try:
+                                send_failure = False
+
+                                if isinstance(exc, MaxNumberOfServersReached):
+                                    send_failure = True
+
+                                if isinstance(exc, APIException):
+                                    if exc.code == "resource_limit_exceeded":
+                                        send_failure = True
+
+                                if send_failure:
+                                    with Action(
+                                        f"Adding scale up failure {exc} message to mailbox for {future.server_name}",
+                                        server_name=future.server_name,
+                                        interval=interval,
+                                    ):
+                                        mailbox.put(
+                                            ScaleUpFailureMessage(
+                                                time=time.time(),
+                                                labels=future.server_labels,
+                                                server_name=future.server_name,
+                                                exception=exc,
+                                            )
+                                        )
+                            finally:
+                                raise
 
             except Exception as exc:
                 msg = f"❗ Error: {type(exc).__name__} {exc}"
